/*
 * util.c:
 * Various utility functions.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: util.c,v 1.3 2001/02/04 18:03:42 chris Exp $";

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* xwrite:
 * Write some data, taking account of short writes and signals.
 */
ssize_t xwrite(int fd, const void *buf, size_t count) {
    size_t c = count;
    const char *b = (const char*)buf;
    while (c > 0) {
        int e = write(fd, b, count);
        if (e > 0) {
            c -= e;
            b += e;
        } else return e;
    } while (c > 0);
    return count;
}

/* xread:
 * Read some data, taking account of signals.
 */
ssize_t xread(int fd, void *buf, size_t count) {
    ssize_t r;
    if (count == 0) return 0;
    do {
        r = read(fd, buf, count);
    } while (r == -1 && errno == EINTR);
    return r;
}

/* vxsprintf:
 * Format some data into a string, which is allocated to be sufficiently large
 * for the data. Also supports %m for "strerror(errno)". Returns NULL on
 * failure, else a string allocated with malloc(3).
 */
#define GUESS 32
char *vxsprintf(const char *fmt, va_list ap) {
    size_t l = GUESS;
    int n;
    char *s, *e = strerror(errno), *t;
    size_t le = strlen(e);
    const char *p, *q;
    
    if (!(s = malloc(l))) return NULL;
    memset(s, 0, l);

    /* First, we need to substitute errors into the string. */
    for (p = fmt, q = strstr(p, "%m"); q; p = q, q = strstr(p, "%m")) {
        while (((q - p) + le) > (l - strlen(s))) {
            l *= 2;
            s = realloc(s, l);
        }
        strncat(s, p, q - p);
        strcat(s, e);
        q += 2;
    }
    while ((fmt + strlen(fmt) - p) > l - strlen(s)) {
        l *= 2;
        if (!(s = realloc(s, l))) return NULL;
    }

    strcat(s, p);

    /* Now, need to use the generated string as a format string to pass to vsnprintf. */
    while (1) {
        t = (char*)malloc(n = l);
        n = vsnprintf(t, n, s, ap);
        if (n > -1 && n < l) {
            free(s);
            return t;
        }
        if (n > -1) l = n + 1;  /* Cope with vsnprintf returning either the required amount of space, or -1 to indicate "need more". */
        else l *= 2;
        if (!(t = realloc(t, l))) return NULL;
    }
}

/* xsprintf:
 * vxsprintf with proper variable argument list.
 */
char *xsprintf(const char *fmt, ...) {
    va_list ap;
    char *s;
    va_start(ap, fmt);
    s = vxsprintf(fmt, ap);
    va_end(ap);
    return s;
}

/* strtonum:
 * Convert a string to an integer, returing 1 if the conversion was
 * successful, or 0 otherwise.
 */
int strtonum(const char *s, int *i) {
    char *p;
    int n;
    n = (int)strtol(s, &p, 10);
    if (p == s) return 0;
    else {
        *i = (int)n;
        return 1;
    }
}
