/*
 * xdata.c:
 * Random xdata stuff.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdata.c,v 1.13 2001/02/15 12:18:26 chris Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/stat.h>

#include "util.h"
#include "xdata.h"

#define IS_INTERNAL(a)      (strncmp((a), "$$:", 3) == 0)

/* xdata_set_error:
 * Set the error status in the xdata object.
 */
void xdata_set_error(XDATA x, const XDATA_ERR status, char *errmsg) {
    if (!x) return;
    if (x->errmsg) free(x->errmsg);
    x->status = status;
    if (errmsg) x->errmsg = errmsg;
    else {
        switch(status) {
        case success: x->errmsg = NULL;
                      break;
        case error:   x->errmsg = strdup("An error occurred");
                      break;
        }
    }
}

/* xdata_open:
 * Start up xdata stuff.
 */
XDATA xdata_open(const char *path) {
    XDATA x;
    int i;
    const char *p;
    char **q, *r;

    if (!path) return NULL;
    
    x = (XDATA)malloc(sizeof(struct tag_XDATA));
    if (!x) return NULL;

    memset(x, 0, sizeof(struct tag_XDATA));

    /* Figure out how many items are in the path. */
    for (i = 1, p = path; *p; ++p)
        if (*p == ':') ++i;

    x->path = (char**)malloc((i + 1) * sizeof(char*));
    x->pdata = strdup(path);

    /* Now copy the path items to the XDATA structure. */
    *(q = x->path) = strtok_r(x->pdata, ":", &r);
    ++q;
    if (x->path[0]) while ((*q++ = strtok_r(NULL, ":", &r)));
    else x->path[0] = x->pdata;

    return x;
}

/* xdata_close:
 * Stop the xdata stuff.
 */
void xdata_close(XDATA x) {
    if (!x) return;
    if (x->path) free(x->path);
    if (x->pdata) free(x->pdata);
    if (x->gdbm_name) free(x->gdbm_name);
    if (x->gdbm_h) gdbm_close(x->gdbm_h);
    free(x);
}

/* xdata_set_type:
 * Set the "type" we are using. Writable is 1 if we want to be able to write
 * to the database; if a database should be created if none yet exists,
 * create_path should be set. Returns 1 on success, or 0 on failure.
 */
int xdata_set_type(XDATA x, const char *type, const int writable, const char *create_path) {
    const char *p;
    char *esctype, *q;
    char **dir;
    
    if (!x) return 0;
    if (x->gdbm_name) free(x->gdbm_name);
    if (x->gdbm_h) gdbm_close(x->gdbm_h);

    /* Now try to find a data file containing the correct type; we must escape
     * the type, since it is allowed to contain /, which is not allowed in a
     * filename.
     */
    esctype = (char*)malloc(strlen(type) * 2 + 1);
    memset(esctype, 0, strlen(type) * 2 + 1);
    for (p = type, q = esctype; *p; ++p) {
        if (strchr("/=", *p)) {
            switch (*p) {
                case '/':   strcpy(q, "=_");
                            break;
                case '=':   strcpy(q, "==");
                            break;
                default:;   /* Can't happen. */
            }
            q += 2;
        } else *q++ = *p;
    }

    for (dir = x->path; *dir; ++dir) {
        char *name = xsprintf("%s/%s", *dir, esctype);
        struct stat st;

        if (stat(name, &st) != -1) {
            int rw = GDBM_READER | (writable ? GDBM_WRITER : 0);
            x->gdbm_h = gdbm_open(name, 512, rw, 0600, NULL);
            if (x->gdbm_h) {
                xdata_set_error(x, success, NULL);
                x->gdbm_name = name;
                x->type = strdup(type);
                xdata_get_metadata(x);
                if (writable) xdata_set_metadata(x);
                break;
            } else
                xdata_set_error(x, error, xsprintf("Data found for type `%s', but file `%s' could not be opened: %s", type, name, gdbm_strerror(gdbm_errno)));
        }
        
        free(name);
    }

    if (!x->gdbm_name) {
        if (create_path) {
            /* OK, need to create a database. */
            char *name = xsprintf("%s/%s", create_path, esctype);
            x->gdbm_h = gdbm_open(name, 512, GDBM_READER | GDBM_WRITER | GDBM_WRCREAT, 0666, NULL);
            if (x->gdbm_h) {
                xdata_set_error(x, success, NULL);
                x->gdbm_name = name;
                x->type = strdup(type);

                x->metadata.last_sync = x->metadata.last_purge = time(NULL);
                xdata_set_metadata(x);
            } else {
                xdata_set_error(x, error, xsprintf("No data found for type `%s', and file `%s' could not be created: %s", type, name, gdbm_strerror(gdbm_errno)));
                free(name);
            }
        }
    }

    free(esctype);

    if (x->gdbm_h) {
        /* Now, it is possible that we need to purge the data in the database. */
        if (writable && (x->metadata.last_purge + XDATA_PURGE_TIME) < time(NULL))
            xdata_purge(x);

        return 1;
    } else {
        if (x->status != error)
            xdata_set_error(x, error, xsprintf("No data found for type `%s'", type));
        return 0;
    }
}

/* xdata_get_metadata:
 * Get the metadata record from the database.
 */
void xdata_get_metadata(XDATA x) {
    void *v;
    size_t len;
    v = xdata_get(x, "$$:metadata", &len, NULL);
    if (!v || len != sizeof(struct xdata_metadata)) {
        /* No metadata; make some up. */
        x->metadata.last_sync = x->metadata.last_purge = time(NULL);
    } else
        /* Copy metadata. */
        memcpy(&x->metadata, v, len);

    free(v);
}

/* xdata_set_metadata:
 * Copy the in-memory metadata into the database.
 */
void xdata_set_metadata(XDATA x) {
    xdata_put(x, "$$:metadata", &x->metadata, sizeof(struct xdata_metadata), time(NULL), 1);
}

/* xdata_purge:
 * Purge stale data from the database. Returns 1 on success or 0 on failure.
 */
int xdata_purge(XDATA x) {
    time_t now = time(NULL);
    char *key, **keys, **k;
    int nkeys, nkeys_a, i;
    int ret = 1;

    key = NULL;
    k = keys = (char**)calloc(nkeys_a = 16, sizeof(char*));
    nkeys = 0;
    while (xdata_enum(x, 1, &key)) {
        if (strncmp(key, "$$:d:", 5) == 0) {        /* Found a deleted-key reference. */
            /* Match. */
            *k++ = strdup(key);
            ++nkeys;
            if (k == keys + nkeys_a) {
                keys = realloc(keys, (nkeys_a *= 2) * sizeof(char*));
                k = keys + nkeys;
            }
        }
    }

    for (i = 0; i < nkeys; ++i) {
        time_t mtime;
        if (!xdata_stat(x, keys[i], NULL, &mtime))
            ret = 0;
        else if ((mtime + XDATA_PURGE_TIME) < now && !xdata_remove(x, keys[i], 0))
            ret = 0;
        free(keys[i]);
    }
    free(keys);

    /* Update metadata. */
    if (ret) {
        x->metadata.last_purge = now;
        xdata_set_metadata(x);
    }
    
    return ret;
}

/* xdata_get:
 * Get a datum from the database. Returns malloc'd data if successful, or NULL
 * if unsuccessful. If successful, the size of the data and its last
 * modification time are copied into *len and *mtime.
 */
void *xdata_get(XDATA x, const char *key, size_t *len, time_t *mtime) {
    void *ret = NULL;
    datum k, v;
    
    if (!x) return NULL;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, strdup("No type has been selected"));
        return NULL;
    }

    k.dptr = (char*)key;
    k.dsize = strlen(key) + 1;
    
    v = gdbm_fetch(x->gdbm_h, k);

    xdata_set_error(x, success, NULL);

    if (v.dptr) {
        if (v.dsize < sizeof(time_t))
            xdata_set_error(x, error, strdup("Stored data is inconsistent"));
        else {
            if (mtime) *mtime = *(time_t*)v.dptr;
            if (len) *len = v.dsize - sizeof(time_t);
            ret = malloc(v.dsize - sizeof(time_t) + 1);
            memcpy(ret, (char*)v.dptr + sizeof(time_t), v.dsize - sizeof(time_t));
        }
    } /* else key not found, which is OK */

    return ret;
}

/* xdata_stat:
 * Get some information about a datum, without retrieving the data itself.
 * Returns 1 on success or 0 on failure.
 */
int xdata_stat(XDATA x, const char *key, size_t *len, time_t *mtime) {
    int ret = 0;
    datum k, v;
    
    if (!x) return 0;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, strdup("No type has been selected."));
        return 0;
    }

    k.dptr = (char*)key;
    k.dsize = strlen(key) + 1;
    
    v = gdbm_fetch(x->gdbm_h, k);

    xdata_set_error(x, success, NULL);

    if (v.dptr) {
        if (v.dsize < sizeof(time_t))
            xdata_set_error(x, error, strdup("Stored data is inconsistent"));
        else {
            if (mtime) *mtime = *(time_t*)v.dptr;
            if (len) *len = v.dsize - sizeof(time_t);
            ret = 1;
        }
    } else xdata_set_error(x, error, strdup("Key not found"));

    return ret;
}

/* xdata_exists:
 * Check for the existence of a datum in the database. Returns 1 if the datum
 * exists, or 0 if it does not.
 */
int xdata_exists(XDATA x, const char *key) {
    datum k;
    int a;
    
    if (!x) return 0;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, strdup("No type has been selected."));
        return 0;
    }

    k.dptr = (char*)key;
    k.dsize = strlen(key) + 1;
    
    a = gdbm_exists(x->gdbm_h, k);

    xdata_set_error(x, success, NULL);

    return a;
}

/* xdata_put:
 * Put a datum into the database. If replace is nonzero, an existing datum
 * will be replaced if present. Returns 1 on success, or 0 on failure.
 */
int xdata_put(XDATA x, const char *key, const void *val, const size_t len, const time_t mtime, const int replace) {
    datum k, v;
    int ret;
    char *key2;

    if (!x) return 0;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, strdup("No type has been selected."));
        return 0;
    }

    k.dptr = (char*)key;
    k.dsize = strlen(key) + 1;
    v.dptr = malloc(len + sizeof(time_t));
    v.dsize = len + sizeof(time_t);
    *(time_t*)v.dptr = mtime;
    memcpy((char*)v.dptr + sizeof(time_t), val, len);

    xdata_set_error(x, success, NULL);
    
    if (gdbm_store(x->gdbm_h, k, v, replace ? GDBM_REPLACE : GDBM_INSERT) == 0)
        ret = 1;
    else {
        xdata_set_error(x, error, xsprintf("Cannot put data for key: %s", gdbm_strerror(gdbm_errno)));
        ret = 0;
    }

    /* At this stage, we may have to remove an old this-key-is-deleted record. */
    key2 = xsprintf("$$:d:%s", key);
    if (!xdata_remove(x, key2, 0)) {
        xdata_set_error(x, error, xsprintf("Cannot remove old deletion record: %s", gdbm_strerror(gdbm_errno)));
        ret = 0;
    }
    free(key2);

    free(v.dptr);

    return ret;
}

/* xdata_remove:
 * Mark a datum as deleted. Returns 1 on success, or 0 on failure.
 */
int xdata_remove(XDATA x, const char *key, const time_t mtime) {
    datum k;
    int ret = 0;

    if (!x) return 0;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, "No type has been selected.");
        return 0;
    }

    k.dptr = (char*)key;
    k.dsize = strlen(key) + 1;

    if (gdbm_exists(x->gdbm_h, k)) {
        if (gdbm_delete(x->gdbm_h, k))
            xdata_set_error(x, error, xsprintf("Cannot delete data for key: %s", gdbm_strerror(gdbm_errno)));
        else if (!IS_INTERNAL(key)) {
            /* If this is not internal data, then mark it as deleted. */
            char d[1], *key2 = xsprintf("$$:d:%s", key);
            ret = xdata_put(x, key2, d, 0, mtime, 1);
            free(key2);
        } else ret = 1;
    } else ret = 1;

    return ret;
}

/* xdata_enum:
 * Enumerate all of the keys in this database. On first invocation, *key
 * should be NULL, and on subsequent calls it will be replaced by a new
 * value. To evaluate only non-internal keys, set internal to 0; for internal
 * keys only, 1; for all keys, 2 or greater.
 */
char *xdata_enum(XDATA x, const int internal, char **key) {
    datum k;
    char *p;
    
    if (!x) return NULL;
    if (!x->gdbm_h) {
        xdata_set_error(x, error, strdup("No type has been selected."));
        return NULL;
    }

    xdata_set_error(x, success, NULL);

    if (!*key) {
        /* Select first key. */
        k = gdbm_firstkey(x->gdbm_h);
        if (internal != 2) while (k.dptr && (internal ? !IS_INTERNAL(k.dptr) : IS_INTERNAL(k.dptr))) {
            /* Discard inappropriate keys. */
            p = k.dptr;
            k = gdbm_nextkey(x->gdbm_h, k);
            free(p);
        }
    } else {
        /* Select next key. */
        k.dptr = *key;
        k.dsize = strlen(*key) + 1;
        
        /* Select first key. */
        k = gdbm_nextkey(x->gdbm_h, k);
        free(*key);
        if (internal != 2) while (k.dptr && (internal ? !IS_INTERNAL(k.dptr) : IS_INTERNAL(k.dptr))) {
            /* Discard inappropriate keys. */
            p = k.dptr;
            k = gdbm_nextkey(x->gdbm_h, k);
            free(p);
        }
    }

    return *key = k.dptr;
}
