/*
 * xdget.c:
 * Get some data from an xdata store.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdget.c,v 1.6 2001/03/25 23:31:32 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>

#include "util.h"
#include "xdata.h"



void usage() {
    fprintf(stderr,
"xdget: get data from an xdata database\n"
"\n"
"usage: xdget [-h] [-p path] [-t type] key [file]\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"Data associated with the given key, if any, will be written to file, or\n"
"standard output if it is not specified.\n"
"\n"
"If any data are retrieved, xdget will exit succesfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

const char optstring[] = "+hp:t:";

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH"), *key, *file;
    int c;
    char *data;
    size_t len;
    int fd;
    XDATA xd;
    int ret = 1; /* 1 indicates no data retrieved. */

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xdget: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xdget: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;
                
            default:
                fprintf(stderr, "xdget: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xdget: no xdata path specified\n");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xdget: no xdata type specified\n");
        usage();
    }
    
    key = argv[optind++];
    if (!key) {
        fprintf(stderr, "xdget: no key specified\n");
        usage();
    }
    
    file = argv[optind];
    
    /* Start xdata. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xdget: cannot open xdata\n");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 0, NULL)) {
        fprintf(stderr, "xdget: %s\n", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }

    /* Try to retrieve the data. */
    if (file && strcmp(file, "/dev/null") == 0) {
        /* Just check for existence. */
        int i = xdata_exists(xd, key);
        if (!i && xd->status == error)
            fprintf(stderr, "xdget: key `%s': %s\n", key, xd->errmsg);
        else if (i)
            ret = 0;
    } else {
        data = xdata_get(xd, key, &len, NULL);
        if (data) {
            ssize_t i;

            if (file) {
                fd = open(file, O_WRONLY | O_CREAT, 0666);
                if (fd == -1) {
                    fprintf(stderr, "xdget: %s: %s\n", file, strerror(errno));
                    exit(1);
                }
            } else {
                file = "standard output";
                fd = 1;
            }

            if ((i = xwrite(fd, data, len)) != len)
                fprintf(stderr, "xdget: %s: %s\n", file, strerror(errno));

            ret = 0; /* We got some data. */
        }
    }

    xdata_close(xd);
    exit(ret);
}
