/*
 * xdrm.c:
 * Get some data from an xdata store.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdrm.c,v 1.8 2001/03/25 23:31:32 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <sys/types.h>

#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>

#include "util.h"
#include "xdata.h"



void usage() {
    fprintf(stderr,
"xdrm: delete keys from an xdata database\n"
"\n"
"usage: xdrm [-h] [-p path] [-t type] [-i] (key | /regexp/) ...\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -i        Names and regular expressions will match case-insensitively.\n"
"\n"
"   -e        Delete internal and metadata keys instead of user keys. Use\n"
"             with caution.\n"
"\n"
"Data associated with the given keys, or keys matching the given regexps, if\n"
"any, will be removed from the database.\n"
"\n"
"If any key is deleted, xdrm will exit succesfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

const char optstring[] = "+hp:t:ie";

struct match {
    char *str;
    regex_t re;
};

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH");
    char *key;
    int c;
    XDATA xd;
    int ret = 1;
    int nmm, i;
    struct match *mm;
    int nocase = 0, internal = 0;
    size_t nkeys, nkeys_a;
    char **keys, **k;

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xdrm: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xdrm: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;

            case 'i':
                nocase = 1;
                break;

            case 'e':
                internal = 1;
                break;
                
            default:
                fprintf(stderr, "xdrm: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xdrm: no xdata path specified\n");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xdrm: no xdata type specified\n");
        usage();
    }
    
    /* Get the things to match. */
    nmm = argc - optind;

    if (nmm <= 0) {
        fprintf(stderr, "xdrm: no keys or regexps specified\n");
        usage();
    } else {
        int j;
        mm = (struct match*)calloc(nmm, sizeof(struct match));
        for (i = 0, j = optind; i < nmm; ++i, ++j) {
            if (*argv[j] == '/' && *(argv[j] + strlen(argv[j]) - 1) == '/') {
                /* Argument is a regular expression. */
                char *x;
                x = strdup(argv[j] + 1);
                *(x + strlen(x) - 1) = 0;
                if ((i = regcomp(&mm[i].re, x, REG_EXTENDED | REG_NOSUB | (nocase ? REG_ICASE : 0)))) {
                    char errbuf[1024];
                    regerror(i, &mm[i].re, errbuf, sizeof(errbuf));
                    fprintf(stderr, "xdrm: %s: %s\n", argv[j], errbuf);
                    regfree(&mm[i].re);
                    exit(1);
                }
                free(x);
            } else mm[i].str = argv[j];
        }
    }

    /* Start xdata. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xdrm: cannot open xdata\n");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 1, NULL)) {
        fprintf(stderr, "xdrm: %s\n", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }

    /* Walk through the database, recording the keys we are going to delete.
     * We have to do this because a loop which deletes keys whilst enumerating
     * may not correctly visit all the keys.
     */
    key = NULL;
    k = keys = (char**)calloc(nkeys_a = 16, sizeof(char*));
    nkeys = 0;
    while (xdata_enum(xd, internal, &key)) {
        for (i = 0; i < nmm; ++i) {
            if ((mm[i].str && (((nocase ? strcasecmp(mm[i].str, key) : strcmp(mm[i].str, key)) == 0)))
                || (!mm[i].str && regexec(&mm[i].re, key, 0, NULL, 0) == 0)) {
                /* Match. */
                *k++ = strdup(key);
                ++nkeys;
                if (k == keys + nkeys_a) {
                    keys = realloc(keys, (nkeys_a *= 2) * sizeof(char*));
                    k = keys + nkeys;
                }
                break;
            }
        }
    }

    /* Actually remove the keys. */
    for (i = 0; i < nkeys; ++i) {
        if (!xdata_remove(xd, keys[i], time(NULL)))
            fprintf(stderr, "xdrm: key `%s': %s\n", keys[i], xd->errmsg);
        else ret = 0;
        free(keys[nkeys]);
    }
    free(keys);
    
    xdata_close(xd);
    exit(ret);
}
