/*
 * xdsync.c:
 * Synchronise two xdata databases.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: xdsync.c,v 1.21 2001/09/15 15:04:45 chris Exp $";
static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved.$";

#include <sys/types.h>

#include <errno.h>
#include <regex.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/utsname.h>

#include "util.h"
#include "xdata.h"

#ifndef XDATA_SYNC_MAX_OFFSET
#   define XDATA_SYNC_MAX_OFFSET        60  /* seconds */
#endif

const int fd_stdin = 0, fd_stdout = 1, fd_stderr = 2;

int verbose;

/* sigchld_handler:
 * Handler for SIGCHLD.
 */
int sigchld_aborts = 1;
sigjmp_buf sigchld_env;
void sigchld_handler(int s) {
    if (sigchld_aborts) siglongjmp(sigchld_env, 0);
}

/* netstring_recv:
 * Receive a netstring from a file descriptor.
 */
char *netstring_recv(int fd, size_t *plen) {
    char lenbuf[16], *p, *data;
    size_t len;
    for (p = lenbuf; p < lenbuf + sizeof(lenbuf); ++p) {
        if (xread(fd, p, 1) == -1) return NULL;
        if (!strchr(":0123456789", *p)) return NULL;
        if (*p == ':') break;
    }
    if (*p != ':') return NULL;
    *p = 0;
    len = (size_t)atoi(lenbuf);
    data = (char*)malloc(len + 1);
    if (!data) return NULL;
    if (xread(fd, data, len) == -1) {
        free(data);
        return NULL;
    }
    if (plen) *plen = len;
    return data;
}

/* netstring_send:
 * Send a netstring to a file descriptor.
 */
int netstring_send(int fd, const char *data, const size_t len) {
    char *s1 = xsprintf("%d:", (int)len);
    int ret = 1;
    if (xwrite(fd, s1, strlen(s1)) == -1 || xwrite(fd, data, len) == -1) ret = 0;
    free(s1);
    return ret;
}


/* Initial protocol phase:
 *
 *          master                  slave
 *
 *      "hello"         ------>
 *                     <------          (time here)
 * ("proceed"|"abort")  ------>
 *
 */
char *protocol_error;

/* master_init:
 * Initial protocol phase, from the master side. Returns 1 on success or 0 on
 * failure.
 */
int master_init(int tx, int rx) {
    char *str;
    time_t there, here;

    /* Send "hello". */
    if (!netstring_send(tx, "hello", 6)) return 0;

    /* Receive the time at the remote machine. */
    if (!(str = netstring_recv(rx, NULL))) return 0;
    here = time(NULL);
    there = (time_t)atoi(str);
    free(str);

    /* Are the two times close enough? */
    if (abs(there - here) > XDATA_SYNC_MAX_OFFSET) {
        protocol_error = "Clock on remote machine is not synchronised to clock here";
        netstring_send(tx, "abort", 6);
        return 0;
    } else if (netstring_send(tx, "proceed", 8))
        return 1;
    else return 0;
}

/* slave_init:
 * Initial protocol phase, from the slave side. Returns 1 on success or 0 on
 * failure.
 */
int slave_init(int tx, int rx) {
    char *str;
    int i;

    /* Receive "hello", hopefully. */
    if (!(str = netstring_recv(rx, NULL))) return 0;
    i = strcmp(str, "hello");
    free(str);
    if (i) return 0;

    /* Now send the time here. */
    str = xsprintf("%d", (int)time(NULL));
    i = netstring_send(tx, str, strlen(str) + 1);
    free(str);
    if (!i) return 0;

    /* Receive "proceed" or "abort". */
    if (!(str = netstring_recv(rx, NULL))) return 0;
    i = strcmp(str, "proceed");
    free(str);
    if (i) {
        protocol_error = "Aborted by peer";
        return 0;
    } else return 1;
}

/* Synchronisation protocol phase:
 *
 *          master                  slave
 *
 *      (nodename)      ------>
 *                     <------          (time of last sync|"never")
 *  - zero or more of - - - - - - - - - - - - - - - - -
 *
 *      "key"           ------>
 *      (key name)      ------>
 *      (key mtime)     ------>
 *
 *                     <------          ("yes"|"no")
 *  [if "yes"]
 *      (key data)      ------>
 *      
 *  - - - - - - - - - - - - - - - - - - - - - - - - - -
 *  
 *      "done"          ------>
 *
 *  Then, the conversation is repeated, with the roles of master and slave
 *  exchanged.
 */
char **keysdone;
size_t nkeysdone, nkeysdone_a;

/* master_sync:
 * Synchronisation protocol phase, from the master side. Returns 1 on success
 * or 0 on failure.
 */
int master_sync(XDATA xd, int tx, int rx) {
    char *key = NULL, *str = NULL, *data = NULL;
    struct utsname uu;
    int never = 0;
    time_t lastsync;
    int ret = 0;

    /* First, we need to find out when we last synchronised with this host. */
    if (uname(&uu) == -1
        || !netstring_send(tx, uu.nodename, strlen(uu.nodename) + 1)) return 0;
    if (verbose) fprintf(stderr, "xdsync: sent nodename `%s'\n", uu.nodename);
    
    str = netstring_recv(rx, NULL);
    if (!str) return 0;
    if (verbose) fprintf(stderr, "xdsync: received last-sync time `%s'\n", str);

    if (strcmp(str, "never") == 0) never = 1;
    else if (!strtonum(str, (int*)&lastsync)) {
        free(str);
        return 0;
    }

    free(str);
    str = NULL;
    
    /* Enumerate all relevant keys. */
    while (xdata_enum(xd, 2, &key)) {
        /* What sort of key is it? We do not send internal keys, except those
         * used to describe deleted data.
         */
        if (strncmp(key, "$$:", 3) != 0 || strncmp(key, "$$:d:", 5) == 0) {
            size_t len;
            time_t mtime;

            data = xdata_get(xd, key, &len, &mtime);
            if (!data) goto fail;

            if (never || mtime > lastsync) {
                int i, f = 0;
                
                /* We may need to send this key. See if it has already been
                 * synchronised this session.
                 */
                if (keysdone) for (i = 0; i < nkeysdone; ++i)
                    if (strcmp(keysdone[i], key) == 0) {
                        f = 1;
                        break;
                    }

                if (!f) {
                    char mtimebuf[16];

                    if (verbose) fprintf(stderr, "xdsync: offering data for key `%s'\n", key);

                    sprintf(mtimebuf, "%d", (int)mtime);

                    if (!netstring_send(tx, "key", 4)
                            || !netstring_send(tx, key, strlen(key) + 1)
                            || !netstring_send(tx, mtimebuf, strlen(mtimebuf) + 1))
                        goto fail;

                    str = netstring_recv(rx, NULL);
                    if (!str) goto fail;

                    if (verbose) fprintf(stderr, "xdsync: received response `%s'\n", str);
                    
                    /* Do we need to send the data? */
                    if (strcmp(str, "yes") == 0 && !netstring_send(tx, data, len)) {
                        goto fail;
                    }
                    free(str);
                    str = NULL;
                    free(data);
                    data = NULL;
                }
            }
        }
    }

    if (!netstring_send(tx, "done", 5)) goto fail;

    if (verbose) fprintf(stderr, "xdsync: completed sync master phase\n");
    ret = 1;

fail:
    if (str) free(str);
    if (key) free(key);
    if (data) free(data);

    return ret;
}

/* slave_sync:
 * Synchronisation protocol phase, from the slave side. Returns 1 on success
 * or 0 on failure.
 */
int slave_sync(XDATA xd, int tx, int rx) {
    char *str = NULL, *metakey = NULL, *key = NULL, *data = NULL;
    time_t lastsync = 0;
    int i, ret = 0;

    if (!keysdone) {
        keysdone = (char**)calloc(nkeysdone_a = 16, sizeof(char*));
        nkeysdone = 0;
    }

    /* Receive a nodename from the remote host. */
    str = netstring_recv(rx, NULL);
    if (!str) return 0;
    if (verbose) fprintf(stderr, "xdsync: received nodename `%s'\n", str);
    metakey = xsprintf("$$:sync:%s", str);
    free(str);
    str = NULL;
    
    /* Find out when we last synchronised with this host, and tell them. */
    if (xdata_stat(xd, metakey, NULL, &lastsync)) {
        char timebuf[16];
        sprintf(timebuf, "%d", (int)lastsync);
        i = netstring_send(tx, timebuf, strlen(timebuf) + 1);
    } else i = netstring_send(tx, "never", 6);

    if (!i) goto fail;
    if (verbose) fprintf(stderr, "xdsync: transmitted last-sync time\n");

    /* Now, receive synchronisation offers from the remote host. */
    for (;;) {
        time_t mtime, mtime_there;
        int dosendback = 0;

        str = netstring_recv(rx, NULL);
        
        if (!str) goto fail;
        if (strcmp(str, "done") == 0) break;

        /* Receive key and modification time. */
        key = netstring_recv(rx, NULL);
        if (!key) goto fail;
        str = netstring_recv(rx, NULL);
        if (!strtonum(str, (int*)&mtime_there)) goto fail;
        free(str);
        str = NULL;

        if (verbose) fprintf(stderr, "xdsync: received offer for key `%s'\n", key);

        /* Stat our version of the key. */
        i = xdata_stat(xd, key, NULL, &mtime);
        if (!i || mtime < mtime_there) {
            size_t len;

            if (strncmp(key, "$$:d:", 5) == 0) {
                /* Special key indicating deletion. */
                if (verbose) fprintf(stderr, "xdsync: interpreted key as a deletion request\n");
                if (!netstring_send(tx, "no", 3))
                    goto fail;

                /* Now, we may already have a more recent copy of this key, in
                 * which case we should ignore the deletion request.
                 */
                if (xdata_exists(xd, key + 5)) {
                    time_t mtime2;
                    if (!xdata_stat(xd, key + 5, NULL, &mtime2)) goto fail;

                    if (mtime2 < mtime_there && !xdata_remove(xd, key + 5, mtime_there)) goto fail;
                    dosendback = 1;
                }
            } else {
                char *key2;
                /* Old or nonexistent key here. */
                if (verbose) fprintf(stderr, "xdsync: accepting new or updated key\n");
                if (!netstring_send(tx, "yes", 4)) goto fail;

                data = netstring_recv(rx, &len);
                if (!data) goto fail;

                if (!xdata_put(xd, key, data, len, mtime_there, 1)) goto fail;
                free(data);
                data = NULL;

                /* Now, it's possible that there exists a deletion request for
                 * this key, too; we should delete that.
                 */
                key2 = xsprintf("$$:d:%s", key);
                if (xdata_exists(xd, key2) && !xdata_remove(xd, key2, mtime_there)) {
                    free(key2);
                    goto fail;
                }
                free(key2);
            }

            /* Add this to a list of keys which we will not transmit back. */
            if (!dosendback) {
                keysdone[nkeysdone++] = strdup(key);
                if (nkeysdone == nkeysdone_a) {
                    nkeysdone_a *= 2;
                    keysdone = (char**)realloc(keysdone, nkeysdone_a * sizeof(char*));
                    if (!keysdone) goto fail;
                }
            }
        } else {
            if (!netstring_send(tx, "no", 3)) goto fail;
            if (verbose) fprintf(stderr, "xdsync: not interested in that key\n");
        }
    }

    /* OK, we have completed this OK, save information about the
     * synchronisation.
     */
    if (!xdata_put(xd, metakey, NULL, 0, time(NULL), 1)) goto fail;

    if (verbose) fprintf(stderr, "xdsync: completed sync slave phase\n");
    ret = 1;

fail:
    if (str) free(str);
    if (metakey) free(metakey);
    if (key) free(key);
    if (data) free(data);

    return ret;
}

/* usage:
 * Print usage information and exit.
 */
void usage() {
    fprintf(stderr,
"xdsync: synchronise two xdata databases\n"
"\n"
"usage: xdsync [-h] [-p path] [-P] [-t type] [-r rsh] [-x xdsync] [-v]\n"
"                   (-s | [user@]host)\n"
"\n"
"   -h        Displays a help message.\n"
"\n"
"   -p path   Specified the path in which to look for xdata databases; if\n"
"             not specified, path is taken from the environment variable\n"
"             XDPATH\n"
"\n"
"   -P        The local xdata path will be passed to the remote machine;\n"
"             this may well be inappropriate, but could be useful if the\n"
"             remote machine is running a broken shell.\n"
"\n"
"   -t type   Specifies the type of data; if not specified, type is taken\n"
"             from the environment variable XDTYPE.\n"
"\n"
"   -r rsh    Specifies the program to be used to contact the remote host,\n"
"             by default, `"XDATA_RSH"'\n"
"\n"
"             If the environment variable XDRSH is set, it will override\n"
"             this value.\n"
"\n"
"   -x xdsync Specifies the name under which to call xdsync on the remote\n"
"             host, by default `xdsync'\n"
"\n"
"   -v        Be verbose.\n"
"\n"
"   -s        Slave mode (internal use only).\n"
"\n"
"xdsync will attempt to synchronise the database for type against that on the\n"
"remote host (optionally, for the given user).\n"
"\n"
"If synchronisation proceeds correctly, then xdsync will return successfully.\n"
"\n"
"Copyright (c) 2001 Chris Lightfoot. Distribute freely.\n");

    exit(1);
}

const char optstring[] = "+hp:Pt:r:x:vs";

int main(int argc, char *argv[]) {
    const char *type = getenv("XDTYPE"), *path = getenv("XDPATH"), *rsh = getenv("XDRSH"), *xdsync = "xdsync";
    char *user, *host;
    int c;
    XDATA xd;
    int ret = 1;
    int slave = 0, passpath = 0;

    if (!rsh) rsh = XDATA_RSH;

    opterr = 0;
    while ((c = getopt(argc, argv, optstring)) != EOF) {
        switch (c) {
            case 'h':
                usage();
                break;
                
            case 'p':
                if (!optarg) {
                    fprintf(stderr, "xdsync: option -p requires an argument\n");
                    usage();
                } else path = optarg;
                break;

            case 'P':
                passpath = 1;
                break;

            case 't':
                if (!optarg) {
                    fprintf(stderr, "xdsync: option -t requires an argument\n");
                    usage();
                } else type = optarg;
                break;

            case 'r':
                if (!optarg) {
                    fprintf(stderr, "xdsync: option -r requires an argument\n");
                    usage();
                } else rsh = optarg;
                break;

            case 'x':
                if (!optarg) {
                    fprintf(stderr, "xdsync: option -x requires an argument\n");
                    usage();
                } else xdsync = optarg;
                break;

            case 'v':
                verbose = 1;
                break;

            case 's':
                slave = 1;
                break;
                
            default:
                fprintf(stderr, "xdsync: unrecognised option -%c\n", optopt);
                usage();
        }
    }

    if (!path) {
        fprintf(stderr, "xdsync%s: no xdata path specified\n", slave ? "-slave" : "");
        usage();
    }

    if (!type) {
        fprintf(stderr, "xdsync: no xdata type specified\n");
        usage();
    }

    /* Obtain remote machine name. */
    if (!slave) {
        user = argv[optind];
        if (!user) {
            fprintf(stderr, "xdsync: no host specified\n");
            usage();
        }
        if ((host = strchr(user, '@'))) {
            *host++ = 0;
        } else {
            host = user;
            user = NULL;
        }
    }
    
    /* Start xdata. */
    xd = xdata_open(path);
    if (!xd) {
        fprintf(stderr, "xdsync%s: cannot open xdata\n", slave ? "-slave" : "");
        exit(1);
    }

    if (!xdata_set_type(xd, type, 1, xd->path[0])) {
        fprintf(stderr, "xdsync%s: %s\n", slave ? "-slave" : "", xd->errmsg);
        xdata_close(xd);
        exit(1);
    }

    if (verbose) {
        fprintf(stderr, "xdsync: type is `%s'\n", type);
        fprintf(stderr, "xdsync: using `%s' for rsh, `%s' for xdsync\n", rsh, xdsync);
        fprintf(stderr, "xdsync: remote host is `%s'\n", host);
        if (user) fprintf(stderr, "xdsync: remote user is `%s'\n", user);
    }

    if (!slave) {
        /* Pipes for talking to the slave--
         *  ps[0] is the slave's standard input
         *  ps[1] is where we write to
         *  pm[0] is where we read from
         *  pm[1] is the slave's standard output
         */
        int ps[2], pm[2];
        struct sigaction sa;

        if (pipe(ps) == -1 || pipe(pm) == -1) {
            fprintf(stderr, "xdsync: pipe: %s\n", strerror(errno));
            goto fail;
        }

        /* Set up SIGCHLD handling. */
        if (sigsetjmp(sigchld_env, 1)) {
            fprintf(stderr, "xdsync: child terminated\n");
            goto fail;
        }
        
        memset(&sa, 0, sizeof(sa));
        sa.sa_handler = sigchld_handler;
        sa.sa_flags   = SA_NOCLDSTOP | SA_RESTART;
        if (sigaction(SIGCHLD, &sa, NULL) == -1) {
            fprintf(stderr, "xdsync: sigaction: %s\n", strerror(errno));
            goto fail;
        }

        /* We should ignore SIGHUP. */
        sa.sa_handler = SIG_IGN;
        sa.sa_flags = SA_RESTART;
        if (sigaction(SIGHUP, &sa, NULL) == -1) {
            fprintf(stderr, "xdsync: sigaction: %s\n", strerror(errno));
            goto fail;
        }
        
        /* Now, we need to fork and execute an appropriate command. */
        switch (fork()) {
            case 0: {
                /* The command is $rsh [-l user] host $xdsync -s [-p path] -t type */
                char *x_argv[16], **a = x_argv;
                close(ps[1]);
                close(pm[0]);

                /* Swap over fds. */
                dup2(ps[0], fd_stdin);
                dup2(pm[1], fd_stdout);

                /* Form the command line. */
                *a++ = (char*)rsh;
                if (user) {
                    *a++ = "-l";
                    *a++ = user;
                }

                *a++ = host;
                *a++ = (char*)xdsync;
                *a++ = "-s";

                if (passpath) {
                    *a++ = "-p";
                    *a++ = (char*)path;
                }

                *a++ = "-t";
                *a++ = (char*)type;

                *a++ = NULL;
                
                if (verbose) {
                    fprintf(stderr, "xdsync: command to execute: ");
                    for (a = x_argv; *a; ++a) fprintf(stderr, "%s ", *a);
                    fprintf(stderr, "\n");
                }

                execvp(rsh, x_argv);
                fprintf(stderr, "xdsync: execvp: %s\n", strerror(errno));
                exit(1);
            }

            case -1:
                fprintf(stderr, "xdsync: fork: %s\n", strerror(errno));
                goto fail;

            default:
                break;
        }

        /* Now, we are the master, and need to start talking. */
        if (!master_init(ps[1], pm[0]) || !master_sync(xd, ps[1], pm[0])) {
            if (protocol_error) fprintf(stderr, "xdsync: %s\n", protocol_error);
            else if (xd->status) fprintf(stderr, "xdsync: %s\n", xd->errmsg);
            else fprintf(stderr, "xdsync: %s\n", strerror(errno));
            goto fail;
        }
        /* Now the roles are reversed. */
        if (!slave_sync(xd, ps[1], pm[0])) {
            if (xd->status) fprintf(stderr, "xdsync: %s\n", xd->errmsg);
            else fprintf(stderr, "xdsync: %s\n", strerror(errno));
            goto fail;
        }

        sigchld_aborts = 0;
    } else {
        /* We are the slave. Listen and obey. */
        if (!slave_init(fd_stdout, fd_stdin) || !slave_sync(xd, fd_stdout, fd_stdin)) {
            if (protocol_error) goto fail;
            else if (xd->status) fprintf(stderr, "xdsync-slave: %s\n", xd->errmsg);
            else fprintf(stderr, "xdsync-slave: %s\n", strerror(errno));
            goto fail;
        }
        /* Now the roles are reversed. */
        if (!master_sync(xd, fd_stdout, fd_stdin)) {
            if (xd->status) fprintf(stderr, "xdsync-slave: %s\n", xd->errmsg);
            else fprintf(stderr, "xdsync-slave: %s\n", strerror(errno));
            goto fail;
        }
        /* Avoid race condition. */
        sleep(1);
    }
    
fail:
    xdata_close(xd);
    exit(ret);
}
